"""Command-line interface to package treem."""

# pylint: disable=too-many-statements

import argparse
import sys

from treem.io import SWC

from treem.commands.check import check
from treem.commands.view import view
from treem.commands.find import find
from treem.commands.modify import modify
from treem.commands.repair import repair
from treem.commands.measure import measure
from treem.commands.convert import convert

try:
    import OpenGL
    from treem.commands.render import render, _HELP
except ImportError:
    pass


def cli():
    """Command-line interface definition."""
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()

    cmd_check = subparsers.add_parser(
        'check',
        epilog='prints out error codes and IDs of error nodes; '
               'returns the number of errors',
        help='test morphology reconstruction for structural consistency')  # noqa
    cmd_check.add_argument('file', type=str,
                           help='input morphology file (swc)')
    cmd_check.add_argument('-q', dest='quiet', action='store_true',
                           help='disable output')
    cmd_check.add_argument('-o', dest='out', metavar='<str>', type=str,
                           help='save output to file (json)')
    cmd_check.set_defaults(func=check)

    cmd_view = subparsers.add_parser('view', help='view morphology')
    cmd_view.add_argument('file', type=str, nargs='+',
                          help='input morphology file (swc)')
    cmd_view.add_argument('-p', dest='type', metavar='<int>', type=int,
                          nargs='+', choices=SWC.TYPES,
                          help='point type {1,2,3,4} [all]')
    cmd_view.add_argument('-b', dest='branch', type=int, metavar='<int>',
                          nargs='+', action='append', help='branch start id')
    cmd_view.add_argument('-s', dest='sec', type=int, metavar='<int>',
                          nargs='+', action='append', help='section start id')
    cmd_view.add_argument('-m', dest='mark', metavar='<int>', type=int,
                          nargs='+', action='append', help='marker point id')
    cmd_view.add_argument('--show-id', dest='show_id', action='store_true',
                          help='show id labels')
    cmd_view.add_argument('-t', dest='title', metavar='<str>', type=str,
                          help='set figure title')
    cmd_view.add_argument('--no-axes', dest='no_axes', action='store_true',
                          help='disable axes')
    cmd_view.add_argument('--scale', dest='scale', metavar='<float>',
                          type=float, help='show scale bars, um')
    cmd_view.add_argument('-c', dest='mode', metavar='<str>', type=str,
                          choices=['neurites', 'cells', 'shadow'],
                          default='neurites',
                          help='color mode {neurites,cells,shadow} '
                               '[neurites]')
    cmd_view.add_argument('--shadow-color', dest='shadow_color',
                          metavar='<str>', type=str, default='lightgray',
                          help='shadow color [lightgray]')
    cmd_view.add_argument('--shadow-width', dest='shadow_width',
                          metavar='<float>', type=float, default=3.0,
                          help='shadow width [3.0]')
    cmd_view.add_argument('-a', dest='angle', metavar='<float>', type=float,
                          nargs=2, help='initial rotation angles, deg. '
                                        '(elevation, azimuth)')
    cmd_view.add_argument('-d', dest='dist', metavar='<float>', type=float,
                          default=10.0,
                          help='initial distance to object, rel. [10.0]')
    cmd_view.add_argument('-j', dest='proj', metavar='<str>', type=str,
                          choices=['xy', 'xz', 'yz'],
                          help='projection {xy,xz,yz}')
    cmd_view.add_argument('-o', dest='out', metavar='<str>', type=str,
                          help='save image to file')
    cmd_view.set_defaults(func=view)

    cmd_find = subparsers.add_parser('find', epilog='prints out point ids',
                                     help='locate single points')
    cmd_find.add_argument('file', type=str,
                          help='input morphology file (swc)')
    cmd_find.add_argument('-p', dest='type', metavar='<int>', type=int,
                          nargs='+', choices=SWC.TYPES,
                          help='point type {1,2,3,4} [any]')
    cmd_find.add_argument('-e', dest='order', metavar='<int>', type=int,
                          nargs='+', help='branch order')
    cmd_find.add_argument('-b', dest='breadth', metavar='<int>', type=int,
                          nargs='+', help='branch breadth')
    cmd_find.add_argument('-g', dest='degree', metavar='<int>', type=int,
                          nargs='+', help='node degree')
    cmd_find.add_argument('-s', dest='stem', action='store_true',
                          help='stem branches (same as: -e 1 --sec -p 2 3 4)')
    cmd_find.add_argument('-d', dest='diam', metavar='<float>', type=float,
                          help='diameter threshold, um')
    cmd_find.add_argument('-l', dest='length', metavar='<float>', type=float,
                          help='segment length threshold, um')
    cmd_find.add_argument('-z', dest='jump', metavar='<float>', type=float,
                          help='z-jump threshold, um')
    cmd_find.add_argument('--comp', dest='compare', metavar='<str>',
                          type=str,
                          choices=['gt', 'lt', 'eq'], default='gt',
                          help='comparison condition for threshold '
                               '{lt,eq,gt} [gt]')
    cmd_find.add_argument('-c', dest='cut', metavar='<float>', type=float,
                          help='cut plane thickness, um (z axis)')
    cmd_find.add_argument('--bottom-up', dest='bottom_up',
                          action='store_true',
                          help='slice surface down (z axis)')
    cmd_find.add_argument('--sec', dest='sec', action='store_true',
                          help='section start ids only')
    cmd_find.set_defaults(func=find)

    cmd_modify = subparsers.add_parser('modify', help='modify morphology')
    cmd_modify.add_argument('file', type=str,
                            help='input morphology file (swc)')
    cmd_modify.add_argument('-p', dest='type', metavar='<int>', type=int,
                            nargs='+',
                            choices=set(SWC.TYPES).difference((SWC.SOMA,)),
                            help='point type {2,3,4} [all]')
    cmd_modify.add_argument('-e', dest='order', metavar='<int>', type=int,
                            nargs='+', help='branch order')
    cmd_modify.add_argument('-b', dest='breadth', metavar='<int>', type=int,
                            nargs='+', help='branch breadth')
    cmd_modify.add_argument('-i', dest='ids', metavar='<int>', type=int,
                            nargs='+', help='input ids')
    cmd_modify.add_argument('-s', dest='scale', metavar='<float>', type=float,
                            nargs=3, help='scaling factors (x,y,z axes)')
    cmd_modify.add_argument('-r', dest='scale_radius', metavar='<float>',
                            type=float, help='radius scaling factor')
    cmd_modify.add_argument('-t', dest='stretch', metavar='<float>',
                            type=float, help='stretching strength, rel.')
    cmd_modify.add_argument('-m', dest='smooth', metavar='<int>', type=int,
                            help='smoothing strength, iter.')
    cmd_modify.add_argument('-j', dest='jitter', metavar='<float>', type=float,
                            help='random coordinate jitter, um (max)')
    cmd_modify.add_argument('--sec', dest='sec', action='store_true',
                            help='apply jitter per section')
    cmd_modify.add_argument('-w', dest='twist', metavar='<float>', type=float,
                            help='random branch twisting, deg. (max)')
    cmd_modify.add_argument('-u', dest='prune', action='store_true',
                            help='prune branches')
    cmd_modify.add_argument('-a', dest='swap', action='store_true',
                            help='swap branches')
    cmd_modify.add_argument('--seed', dest='seed', metavar='<int>', type=int,
                            help='random seed')
    cmd_modify.add_argument('-o', dest='out', metavar='<str>', type=str,
                            default='mod.swc',
                            help='output morphology file (swc) [mod.swc]')
    cmd_modify.set_defaults(func=modify)

    cmd_repair = subparsers.add_parser('repair', help='repair morphology')
    cmd_repair.add_argument('file', type=str,
                            help='input morphology file (swc)')
    cmd_repair.add_argument('-n', dest='center', action='store_true',
                            help='center root')
    cmd_repair.add_argument('-t', dest='translate', metavar='<float>',
                            type=float, nargs=3,
                            help='translate morphology, um (x,y,z)')
    cmd_repair.add_argument('-a', dest='rotate', metavar='<float>', type=float,
                            nargs=3, help='rotate morphology, deg. '
                                          '(x,y,z axes)')
    cmd_repair.add_argument('-k', dest='shrink', metavar='<float>', type=float,
                            help='shrinkage correction factor (z axis)')
    cmd_repair.add_argument('-kxy', dest='shrink_xy', metavar='<float>',
                            type=float,
                            help='shrinkage correction factor (x,y plane)')
    cmd_repair.add_argument('--bottom-up', dest='bottom_up',
                            action='store_true',
                            help='slice surface down (z axis)')
    cmd_repair.add_argument('--seed', dest='seed', metavar='<int>', type=int,
                            help='random seed')
    cmd_repair.add_argument('-c', dest='cut', metavar='<int>', type=int,
                            nargs='+', help='cut points ids')
    cmd_repair.add_argument('--force-repair', dest='force_repair',
                            action='store_true',
                            help='repair cuts using branches of any order')
    cmd_repair.add_argument('-d', dest='diam', metavar='<int>', type=int,
                            nargs='+',
                            help='set diameter in points ids')
    cmd_repair.add_argument('--diam', dest='diam_mode', metavar='<str>',
                            type=str, default='joint',
                            choices=['joint', 'sec', 'order', 'breadth'],
                            help='correction mode '
                                 '{joint,sec,order,breadth} [joint]')
    cmd_repair.add_argument('--pool', dest='pool', metavar='<str>', type=str,
                            nargs='+', help='reference reconstructions for repair')
    cmd_repair.add_argument('-l', dest='delete', metavar='<int>', type=int,
                            nargs='+',
                            help='delete nodes in points ids')
    cmd_repair.add_argument('-z', dest='zjump', metavar='<int>', type=int,
                            nargs='+', help='z-jump points ids')
    cmd_repair.add_argument('--zjump', dest='zjump_mode', metavar='<str>',
                            type=str, default='align',
                            choices=['align', 'split', 'tilt', 'join'],
                            help='correction mode '
                                 '{align,split,tilt,join} [align]')
    cmd_repair.add_argument('-r', dest='res', metavar='<float>', type=float,
                            help='sampling resolution, um')
    cmd_repair.add_argument('-o', dest='out', metavar='<str>', type=str,
                            default='rep.swc',
                            help='output morphology file (swc) [rep.swc]')
    cmd_repair.add_argument('-v', dest='verbose', action='store_true',
                            help='verbose output')
    cmd_repair.set_defaults(func=repair)

    cmd_measure = subparsers.add_parser('measure', help='measure morphology')
    cmd_measure.add_argument('file', type=str, nargs='+',
                             help='input morphology file (swc)')
    cmd_measure.add_argument('-p', dest='type', metavar='<int>', type=int,
                             nargs='+', choices=SWC.TYPES,
                             help='point type {1,2,3,4} [all]')
    cmd_measure.add_argument('-o', dest='out', metavar='<str>', type=str,
                             help='output morphometric file (json)')
    cmd_measure.set_defaults(func=measure)

    cmd_convert = subparsers.add_parser('convert', help='convert input file')
    cmd_convert.add_argument('file', type=str,
                             help='input morphology file (swc)')
    cmd_convert.add_argument('-p', dest='type', metavar='<int>', type=int,
                             nargs='+', choices=SWC.TYPES,
                             help='point type {1,2,3,4} [all]')
    cmd_convert.add_argument('-o', dest='out', metavar='<str>', type=str,
                             default='inp.swc',
                             help='converted morphology file (swc) [inp.swc]')
    cmd_convert.add_argument('-q', dest='quiet', action='store_true',
                             help='disable output')
    cmd_convert.set_defaults(func=convert)

    if 'OpenGL' in sys.modules:
        cmd_render = subparsers.add_parser(
            'render', help='show 3D model', epilog=_HELP,
            formatter_class=argparse.RawDescriptionHelpFormatter)
        cmd_render.add_argument('file', type=str,
                                help='input morphology file (swc)')
        cmd_render.set_defaults(func=render)

    args = parser.parse_args()
    sys.exit(args.func(args))



"""Testing CLI command view."""

import subprocess
import os


def test_png():
    """Tests for common plot."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'view', 'pass_simple_branch.swc',
                             '-o', '/tmp/test_treem.png'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''


def test_pdf():
    """Tests for plot options."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'view', 'pass_simple_branch.swc',
                             '-t', 'title', '--no-axes', '--show-id',
                             '--scale', '100', '-c', 'cells',
                             '-b', '2', '-s', '4', '-m', '3', '9',
                             '-a', '20', '30',
                             '-o', '/tmp/test_treem.pdf'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''


def test_shadow():
    """Tests for plot options."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'view', 'pass_simple_branch.swc',
                             'pass_zjump.swc', '-c', 'shadow',
                             '-o', '/tmp/test_treem.png'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''


def test_proj_xy():
    """Tests for plot projection."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'view', 'pass_simple_branch.swc',
                             '-j', 'xy',
                             '-o', '/tmp/test_treem.png'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''


def test_proj_xz():
    """Tests for plot projection."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'view', 'pass_simple_branch.swc',
                             '-j', 'xz',
                             '-o', '/tmp/test_treem.png'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''


def test_proj_yz():
    """Tests for plot projection."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'view', 'pass_simple_branch.swc',
                             '-j', 'yz',
                             '-o', '/tmp/test_treem.png'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''



"""Testing module geom."""

import numpy as np

from treem.utils.geom import rotation


def test_tree_rotation():
    """Tests 3D rotation."""
    vec = [1, 0, 0]
    assert np.allclose(rotation(vec, vec)[0],
                       np.array([0., 0.70710678, -0.70710678]))
    assert rotation(vec, vec)[1] == 0.0



"""Testing module tree."""

from treem import Tree


def test_tree_traversal():
    """Tests tree traversal."""
    tree0 = Tree()
    tree1 = tree0.add(Tree())
    tree2 = tree1.add(Tree())
    tree3 = tree1.add(Tree())
    pre = list(x for x in tree0.preorder())
    post = list(x for x in tree0.postorder())
    level = list(x for x in tree0.levelorder())
    assert pre == [tree0, tree1, tree2, tree3]
    assert post == [tree2, tree3, tree1, tree0]
    assert level == [tree3, tree2, tree1, tree0]
    assert tree0.depth() == 0
    assert tree1.depth() == 1
    assert tree2.depth() == 2
    assert tree3.depth() == 2
    assert tree2.width() == 2
    assert tree0.height() == 2



"""Implementation of CLI render command."""

import numpy as np

from treem import Morph, SWC
from treem.utils.geom import rotation

# pylint: disable=wildcard-import
# pylint: disable=unused-wildcard-import
# pylint: disable=redefined-builtin

from OpenGL.GL import *
from OpenGL.GLU import *
from OpenGL.GLUT import *

from PIL import Image, ImageOps

_HELP = """
interactive commands:
    mouse left drag         rotate
    mouse right drag        translate

    Z/z     zoom in/out
    w       white background
    b       black background
    d       dark background
    p       show/hide points
    l       show/hide lines
    n       show/hide nodes
    s       show/hide segments
    W       write image to file
    h/?     help message
"""

class InteractionMatrix():
    """Class for object rotation and translation."""
    # pylint: disable=invalid-name
    def __init__(self):
        self.current_matrix = None
        self.reset()

    def reset(self):
        """Reset current matrix."""
        glPushMatrix()
        glLoadIdentity()
        self.current_matrix = glGetFloatv(GL_MODELVIEW_MATRIX)
        glPopMatrix()

    def add_translation(self, x, y, z):
        """Shift scene by a vector."""
        glPushMatrix()
        glLoadIdentity()
        glTranslatef(x, y, z)
        glMultMatrixf(self.current_matrix)
        self.current_matrix = glGetFloatv(GL_MODELVIEW_MATRIX)
        glPopMatrix()

    def add_rotation(self, angle, x, y, z):
        """Rotate scene around a vector."""
        glPushMatrix()
        glLoadIdentity()
        glRotatef(angle, x, y, z)
        glMultMatrixf(self.current_matrix)
        self.current_matrix = glGetFloatv(GL_MODELVIEW_MATRIX)
        glPopMatrix()

    def get_current_matrix(self):
        """Return current matrix."""
        return self.current_matrix


class MouseInteractor():
    """Class for mouse control."""
    # pylint: disable=invalid-name
    def __init__(self, translationScale=0.1, rotationScale=0.2):
        self.scaling_factor_rotation = rotationScale
        self.scaling_factor_translation = translationScale
        self.rotation_matrix = InteractionMatrix()
        self.translation_matrix = InteractionMatrix()
        self.mouse_button_pressed = None
        self.old_mouse_pos = [0, 0]
        glutMouseFunc(self.mouse_button)
        glutMotionFunc(self.mouse_motion)

    def mouse_button(self, button, mode, x, y):
        """Detect button press."""
        if mode == GLUT_DOWN:
            self.mouse_button_pressed = button
        else:
            self.mouse_button_pressed = None
        self.old_mouse_pos[0], self.old_mouse_pos[1] = x, y
        glutPostRedisplay()

    def mouse_motion(self, x, y):
        """Detect mouse motion."""
        dx = x - self.old_mouse_pos[0]
        dy = y - self.old_mouse_pos[1]
        if self.mouse_button_pressed == GLUT_RIGHT_BUTTON:
            tx = dx * self.scaling_factor_translation
            ty = dy * self.scaling_factor_translation
            self.translation_matrix.add_translation(tx, -ty, 0)
        elif self.mouse_button_pressed == GLUT_LEFT_BUTTON:
            ry = dx * self.scaling_factor_rotation
            self.rotation_matrix.add_rotation(ry, 0, 1, 0)
            rx = dy * self.scaling_factor_rotation
            self.rotation_matrix.add_rotation(rx, 1, 0, 0)
        else:
            tz = dy * self.scaling_factor_translation
            self.translation_matrix.add_translation(0, 0, tz)
        self.old_mouse_pos[0], self.old_mouse_pos[1] = x, y
        glutPostRedisplay()

    def apply(self):
        """Apply matrix manipulations."""
        glMultMatrixf(self.translation_matrix.get_current_matrix())
        glMultMatrixf(self.rotation_matrix.get_current_matrix())


class App:
    """Interactive application."""
    # pylint: disable=invalid-name
    # pylint: disable=too-many-instance-attributes
    def __init__(self, morph, title='render'):
        glutInit()
        glutInitDisplayMode(GLUT_SINGLE | GLUT_RGB | GLUT_DEPTH)
        glutInitWindowSize(800, 600)
        glutInitWindowPosition(100, 100)
        glutCreateWindow(title)
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        glMatrixMode(GL_MODELVIEW)

        glutDisplayFunc(self.display)
        glutKeyboardFunc(self.keyboard)

        morph.data[:, SWC.XYZ] -= morph.root.coord()
        self.cmin = morph.data[:, 2:5].min(axis=0)
        self.cmax = morph.data[:, 2:5].max(axis=0)

        self.mouse = MouseInteractor(0.1, 1.0)
        self.center = morph.root.coord()
        self.view_size = np.linalg.norm(self.cmax - self.cmin) / 4
        self.camera_dist = self.view_size
        self.camera_tilt = 30.0
        self.camera_rot = 30.0
        self.img_count = 0

        self.color = {
            'dark': (0.3, 0.3, 0.3),
            'white': (1.0, 1.0, 1.0),
            'black': (0.0, 0.0, 0.0),
            'axon': (0.6, 0.6, 0.6),
            'dend': (0.3, 0.5, 0.9),
            'apic': (1.0, 0.4, 0.6),
            'soma': (0.1, 0.3, 0.7),
        }
        red, green, blue = self.color['dark']
        glClearColor(red, green, blue, 0)

        light_ambient = [0.5, 0.5, 0.5, 1.0]
        light_diffuse = [0.8, 0.8, 0.8, 1.0]
        light_position = [self.cmax[0], self.cmax[1], self.cmax[2], 0.0]

        glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient)
        glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse)
        glLightfv(GL_LIGHT0, GL_POSITION, light_position)

        glEnable(GL_LIGHTING)
        glEnable(GL_LIGHT0)
        glEnable(GL_DEPTH_TEST)

        glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE)
        glEnable(GL_COLOR_MATERIAL)
        glEnable(GL_NORMALIZE)

        self.disp_points = self.genlist_points(morph)
        self.disp_lines = self.genlist_lines(morph)
        self.disp_nodes = self.genlist_nodes(morph)
        self.disp_segments = self.genlist_segments(morph)

        self.show_points = False
        self.show_lines = True
        self.show_nodes = True
        self.show_segments = True

    def set_nrncolor(self, objtype):
        """Change current color."""
        if objtype == 1:
            glColor(self.color['soma'])
        elif objtype == 2:
            glColor(self.color['axon'])
        elif objtype == 3:
            glColor(self.color['dend'])
        elif objtype == 4:
            glColor(self.color['apic'])
        else:
            glColor3f(0.5, 0.5, 0.5)

    def genlist_points(self, morph):
        """Create display list for points."""
        disp_list = glGenLists(1)
        glNewList(disp_list, GL_COMPILE)

        glBegin(GL_POINTS)
        self.set_nrncolor(morph.root.type())
        glVertex(morph.root.coord())
        glEnd()

        for stem in morph.stems():
            glBegin(GL_POINTS)
            self.set_nrncolor(stem.type())
            for node in stem.walk():
                glVertex(node.coord())
            glEnd()

        glEndList()
        glPointSize(3)
        return disp_list

    def genlist_lines(self, morph):
        """Create display list for lines."""
        disp_list = glGenLists(1)
        glNewList(disp_list, GL_COMPILE)

        for stem in morph.stems():
            self.set_nrncolor(stem.type())
            for node in stem.walk():
                glBegin(GL_LINES)
                glVertex(node.parent.coord())
                glVertex(node.coord())
                glEnd()

        glEndList()
        return disp_list

    def genlist_nodes(self, morph):
        """Create display list for nodes."""
        disp_list = glGenLists(1)
        glNewList(disp_list, GL_COMPILE)

        self.set_nrncolor(morph.root.type())
        nx, ny, nz = morph.root.coord()
        nr = morph.root.radius()
        glPushMatrix()
        glTranslatef(nx, ny, nz)
        glutSolidSphere(nr, 8, 8)
        glPopMatrix()

        for stem in morph.stems():
            self.set_nrncolor(stem.type())
            for node in stem.walk():
                nx, ny, nz = node.coord()
                nr = node.radius() * 0.95
                glPushMatrix()
                glTranslatef(nx, ny, nz)
                glutSolidSphere(nr, 8, 8)
                glPopMatrix()

        glEndList()
        return disp_list

    def genlist_segments(self, morph):
        """Create display list for segments."""
        disp_list = glGenLists(1)
        glNewList(disp_list, GL_COMPILE)

        for stem in morph.stems():
            self.set_nrncolor(stem.type())
            for node in stem.walk():
                nr = node.radius()
                px, py, pz = node.parent.coord()
                pr = node.parent.radius()
                pr /= 2.0 if node.parent == morph.root else 1.0
                glPushMatrix()
                glTranslatef(px, py, pz)
                axis, angle = rotation(
                    (0.0, 0.0, 1.0), node.coord() - node.parent.coord())
                rx, ry, rz = axis
                glRotate(angle / np.pi * 180, rx, ry, rz)
                gluCylinder(gluNewQuadric(), pr, nr, node.length(), 8, 1)
                glPopMatrix()

        glEndList()
        return disp_list

    def display(self):
        """Display callback function."""
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        width, height = glGetFloatv(GL_VIEWPORT)[2:4]
        aspect = width / height
        total_size = np.linalg.norm(self.cmax - self.cmin)
        glOrtho((self.center[0] - self.view_size) * aspect,
                (self.center[0] + self.view_size) * aspect,
                self.center[1] - self.view_size,
                self.center[1] + self.view_size,
                -total_size, total_size)
        glTranslate(0.0, 0.0, -self.camera_dist)
        glRotate(self.camera_tilt, 1.0, 0.0, 0.0)
        glRotate(self.camera_rot, 0.0, 1.0, 0.0)
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()
        self.mouse.apply()

        if self.show_points:
            glCallList(self.disp_points)
        if self.show_lines:
            glCallList(self.disp_lines)
        if self.show_nodes:
            glCallList(self.disp_nodes)
        if self.show_segments:
            glCallList(self.disp_segments)

        glFlush()

    def write_img(self):
        """Save image buffer to file."""
        # pylint: disable=no-value-for-parameter
        glPixelStorei(GL_PACK_ALIGNMENT, 1)
        width, height = glGetFloatv(GL_VIEWPORT)[2:4]
        data = glReadPixels(0, 0, width, height, GL_RGBA, GL_UNSIGNED_BYTE)
        image = Image.frombytes('RGBA', (width, height), data)
        image = ImageOps.flip(image)
        self.img_count += 1
        image.save(f'render_{self.img_count}.png', 'PNG')

    def keyboard(self, key, x, y):  # pylint: disable=unused-argument
        """Keyboard callback function."""
        key = key.decode('UTF-8')
        if key == 'z':
            self.view_size *= 1.1
        elif key == 'Z':
            self.view_size *= 0.9
        elif key == 'w':
            red, green, blue = self.color['white']
            glClearColor(red, green, blue, 0)
        elif key == 'b':
            red, green, blue = self.color['black']
            glClearColor(red, green, blue, 0)
        elif key == 'd':
            red, green, blue = self.color['dark']
            glClearColor(red, green, blue, 0)
        elif key == 'p':
            self.show_points = not self.show_points
        elif key == 'l':
            self.show_lines = not self.show_lines
        elif key == 'n':
            self.show_nodes = not self.show_nodes
        elif key == 's':
            self.show_segments = not self.show_segments
        elif key == 'W':
            self.write_img()
        elif key in ('h', '?'):
            print(_HELP)
        glutPostRedisplay()

    def run(self):
        """Enter event processing loop."""
        # pylint: disable=no-self-use
        glutMainLoop()



def render(args):
    """Shows 3D model of morphology reconstruction."""
    morph = Morph(args.file)
    app = App(morph, title=args.file)
    app.run()



"""Testing CLI command find."""

import subprocess
import os


def test_all_nodes():
    """Tests for returning all nodes."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'find', 'pass_simple_branch.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == '1 2 3 4 5 6 7 8 9 10 11 12 13 \n'
    assert stderr == ''


def test_order():
    """Tests for nodes of the given branch order."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'find', 'pass_simple_branch.swc',
                             '-e', '2'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == '3 4 5 6 7 8 \n'
    assert stderr == ''


def test_order_sec():
    """Tests for sections of the given branch order."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'find', 'pass_simple_branch.swc',
                             '-e', '2', '--sec'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == '4 8 \n'
    assert stderr == ''


def test_breadth():
    """Tests for nodes of the given branch breadth."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'find', 'pass_simple_branch.swc',
                             '-b', '1'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == '4 5 6 7 10 11 12 13 \n'
    assert stderr == ''


def test_degree():
    """Tests for nodes of the given degree."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'find', 'pass_simple_branch.swc',
                             '-g', '2'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == '3 9 \n'
    assert stderr == ''


def test_stem():
    """Tests for stem starting nodes."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'find', 'pass_simple_branch.swc',
                             '-s'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == '2 \n'
    assert stderr == ''


def test_diam_gt():
    """Tests for diameter values."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'find', 'pass_simple_branch.swc',
                             '-d', '1'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == '1 \n'
    assert stderr == ''


def test_diam_lt():
    """Tests for diameter values."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'find', 'pass_simple_branch.swc',
                             '-d', '0.1', '--comp', 'lt'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == '5 \n'
    assert stderr == ''


def test_diam_eq():
    """Tests for diameter values."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'find', 'pass_simple_branch.swc',
                             '-d', '0.04', '--comp', 'eq'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == '5 \n'
    assert stderr == ''


def test_length_gt():
    """Tests for length values."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'find', 'pass_zjump.swc',
                             '-l', '25'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == '2 10 11 12 \n'
    assert stderr == ''


def test_length_lt():
    """Tests for length values."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'find', 'pass_zjump.swc',
                             '-l', '25', '--comp', 'lt'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == '1 3 4 5 6 7 8 9 \n'
    assert stderr == ''


def test_length_eq():
    """Tests for length values."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'find', 'pass_zjump.swc',
                             '-l', '0.0', '--comp', 'eq'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == '1 \n'
    assert stderr == ''


def test_zjump_gt():
    """Tests for zjump values."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'find', 'pass_zjump.swc',
                             '-z', '19'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == '4 \n'
    assert stderr == ''


def test_zjump_lt():
    """Tests for zjump values."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'find', 'pass_zjump.swc',
                             '-z', '19', '--comp', 'lt'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == '2 3 5 6 7 8 9 10 11 12 \n'
    assert stderr == ''


def test_zjump_eq():
    """Tests for zjump values."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'find', 'pass_zjump.swc',
                             '-z', '20', '--comp', 'eq'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == '4 \n'
    assert stderr == ''


def test_cut():
    """Tests for cut neurites."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'find', 'pass_nmo_2_cut.swc',
                             '-c', '10', '-p', '3'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == '322 341 547 1167 \n'
    assert stderr == ''


def test_cut_bottom_up():
    """Tests for cut neurites."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'find', 'pass_nmo_3_cut.swc',
                             '-c', '10', '-p', '3', '--bottom-up'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == '521 1764 1848 1938 1956 1978 \n'
    assert stderr == ''



"""Testing CLI command check."""

import subprocess
import os


def test_no_file():
    """Tests for missing file."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'check', 'fail_no_file.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 1
    assert stdout == 'no_file: fail_no_file.swc\n'
    assert stderr == ''


def test_no_data():
    """Tests for empty file."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'check', 'fail_no_data.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 1
    assert stdout == 'no_data: True\n'
    assert stderr == ''


def test_node1_has_parent():
    """Tests whether first node is root."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'check', 'fail_node1_has_parent.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 2
    assert stdout == """node1_has_parent: 2
not_valid_parent_ids: 2
"""
    assert stderr == ''


def test_node1_not_id1():
    """Tests whether first node has ID of 1."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'check', 'fail_node1_not_id1.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 2
    assert stdout == """node1_not_id1: 2
non_stem_neurite: 3
"""
    assert stderr == ''


def test_node1_not_soma():
    """Tests whether first node has TYPE of 1."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'check', 'fail_node1_not_soma.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 2
    assert stdout == """node1_not_soma: 3
non_stem_neurite: 1
"""
    assert stderr == ''


def test_non_descendant():
    """Tests whether node ID is greater than parent ID."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'check', 'fail_non_descendant.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 1
    assert stdout == 'non_descendant: 2 3\n'
    assert stderr == ''


def test_non_increasing_ids():
    """Tests whether IDs are increasing."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'check', 'fail_non_increasing_ids.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 1
    assert stdout == 'non_increasing_ids: 3 2\n'
    assert stderr == ''


def test_non_sequential_ids():
    """Tests whether ID increment is 1."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'check', 'fail_non_sequential_ids.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 1
    assert stdout == 'non_sequential_ids: 4 6\n'
    assert stderr == ''


def test_non_stem_neurite():
    """Tests for neurites changing type (e.g., axon from dend)."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'check', 'fail_non_stem_neurite.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 1
    assert stdout == 'non_stem_neurite: 4\n'
    assert stderr == ''


def test_non_unique_ids():
    """Tests for unique IDs."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'check', 'fail_non_unique_ids.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 1
    assert stdout == 'non_unique_ids: 2 3\n'
    assert stderr == ''


def test_not_array_1():
    """Tests for matrix shape of the data."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'check', 'fail_not_array_1.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 1
    assert stdout == 'not_array: True\n'
    assert stderr == ''


def test_not_array_2():
    """Tests for data matrix contains numbers."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'check', 'fail_not_array_2.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 1
    assert stdout == 'not_array: True\n'
    assert stderr == ''


def test_not_swc_cols():
    """Tests for data matrix has shape (N, 7)."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'check', 'fail_not_swc_cols.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 1
    assert stdout == 'not_swc_cols: 8\n'
    assert stderr == ''


def test_not_swc_ext():
    """Tests for data file name extension SWC."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'check', 'fail_not_swc_ext.txt'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 1
    assert stdout == 'not_swc_ext: txt\n'
    assert stderr == ''


def test_not_valid_ids():
    """Tests for valid IDs."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'check', 'fail_not_valid_ids.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 1
    assert stdout == 'not_valid_ids: 0 -1\n'
    assert stderr == ''


def test_not_valid_parent_ids():
    """Tests for valid parent IDs."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'check', 'fail_not_valid_parent_ids.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 1
    assert stdout == 'not_valid_parent_ids: 4 5\n'
    assert stderr == ''


def test_not_valid_types():
    """Tests for valid point types."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'check', 'fail_not_valid_types.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 1
    assert stdout == 'not_valid_types: 6 5\n'
    assert stderr == ''


def test_single_point():
    """Tests for the number of data points greater than 1."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'check', 'fail_single_point.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 1
    assert stdout == 'single_point: 1\n'
    assert stderr == ''


def test_undef_parent_ids():
    """Tests for disconnected nodes."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'check', 'fail_undef_parent_ids.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 1
    assert stdout == 'undef_parent_ids: 3 4\n'
    assert stderr == ''


def test_unordered():
    """Tests unordered SWC file."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'check', 'fail_unordered.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 4
    assert stdout == """node1_not_id1: 2
node1_has_parent: 1
node1_not_soma: 3
not_valid_parent_ids: 1
"""
    assert stderr == ''


def test_simple_branch():
    """Tests unordered SWC file."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'check', 'pass_simple_branch.swc',
                             '-o', '/tmp/test_treem.json'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''



"""Testing module io."""

import json

import numpy as np

from treem.io import TreemEncoder


def test_io_encoder():
    """Tests JSON encoder."""
    data = [[1, 2, 3], np.array(range(4, 7)), {7, 8, 9}, 0, "a"]
    assert json.dumps(data, cls=TreemEncoder) ==\
        '[[1, 2, 3], [4, 5, 6], [8, 9, 7], 0, "a"]'



"""Plotting utilities."""

import numpy as np

from treem.io import SWC

# pylint: disable=invalid-name


def plot_tree(ax, tree, data, **kwargs):
    """Plots entire branch.

    Args:
        ax: matplotlib axes object.
        tree (treem.Node): branch start node.
        data (NumPy ndarray): raw data of morphology Morph.
        kwargs: arguments for matplotlib plot().
    """
    # pylint: disable=too-many-locals
    for sec in tree.sections():
        first = sec[0].ident() - 1
        last = sec[-1].ident()
        block = slice(first, last)
        x, y, z = data[block, SWC.XYZ].T
        ax.plot(x, y, z, **kwargs)
    for bif in tree.forks():
        b = bif.coord()
        for child in bif.siblings:
            c = child.coord()
            a = np.array([b, c])
            x, y, z = a.T  # pylint: disable=unpacking-non-sequence
            ax.plot(x, y, z, **kwargs)


def plot_section(ax, tree, data, **kwargs):
    """Plots single section.

    Args:
        ax: matplotlib axes object.
        tree (treem.Node): branch start node.
        data (NumPy ndarray): raw data of morphology Morph.
        kwargs: arguments for matplotlib plot().
    """
    for sec in tree.sections():
        first = sec[0].ident() - 1
        last = sec[-1].ident()
        block = slice(first, last)
        x, y, z = data[block, SWC.XYZ].T
        ax.plot(x, y, z, **kwargs)
        break


def plot_neuron(ax, morph, types=SWC.TYPES, colors=None, linewidth=1):
    """Plots neuron morphology.

    Args:
        ax: matplotlib axes object.
        morph (treem.Morph): neuron morphology.
        types (int iterable): point types to be displayed.
        colors (dict): colors for point types ({pt: colspec}).
        linewidth (int): line width.
    """
    # pylint: disable=unpacking-non-sequence
    colors = colors if colors else {t: f'C{t}' for t in types}
    r = morph.root.coord()
    for stem in morph.stems():
        if stem.type() in types:
            x, y, z = np.array([r, stem.coord()]).T
            ax.plot(x, y, z, c=colors[stem.type()], lw=linewidth)
            plot_tree(ax, stem, morph.data, c=colors[stem.type()],
                      lw=linewidth)
    if SWC.SOMA in types:
        soma_points = morph.data[np.where(morph.data[:, SWC.T] == SWC.SOMA)]
        x, y, z = soma_points[:, SWC.XYZ].T
        ax.plot(x, y, z, linestyle='', marker='o', markersize=10,
                c=colors[SWC.SOMA], alpha=0.25)
    x, y, z = r
    ax.plot([x], [y], [z], linestyle='', marker='o', markersize=5,
            color='black')


def plot_points(ax, morph, ids, types=SWC.TYPES, show_id=False):
    """Plots marker points.

    Args:
        ax: matplotlib axes object.
        morph (treem.Morph): neuron morphology.
        ids (int iterable): list of node IDs.
        types (int iterable): point types to display.
    """
    points = np.array([morph.data[x-1] for x in ids
                       if morph.data[x-1][SWC.T] in types])
    x, y, z = points[:, SWC.XYZ].T
    ax.plot(x, y, z, linestyle='', marker='.')
    if show_id:
        for point in points:
            x, y, z = point[SWC.XYZ]
            i = point[SWC.I].astype(int)
            ax.text3D(x, y, z, f'  {i}')



"""Testing CLI command repair."""

import subprocess
import os


def test_transpose():
    """Tests for changing location."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'repair', 'pass_simple_branch.swc',
                             '-t', '2', '2', '2',
                             '-a', '90', '90', '90',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''


def test_shrink():
    """Tests for shrinkage correction."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'repair', 'pass_zjump.swc',
                             '-k', '1.33', '-kxy', '1.11', '-n',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''


def test_shrink_bottom_up():
    """Tests for shrinkage correction."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'repair', 'pass_zjump.swc',
                             '-k', '1.33', '-kxy', '1.11',
                             '--bottom-up',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''


def test_zjump_align():
    """Tests for zjump correction."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'repair', 'pass_zjump.swc',
                             '-z', '4', '--zjump', 'align',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''


def test_zjump_split():
    """Tests for zjump correction."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'repair', 'pass_zjump.swc',
                             '-z', '4', '--zjump', 'split',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''


def test_zjump_tilt():
    """Tests for zjump correction."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'repair', 'pass_zjump.swc',
                             '-z', '4', '--zjump', 'tilt',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''


def test_zjump_join():
    """Tests for zjump correction."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'repair', 'pass_zjump.swc',
                             '-z', '4', '--zjump', 'join',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''


def test_diam_joint():
    """Tests for diameter correction."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'repair', 'pass_simple_branch.swc',
                             '-d', '5', '--diam', 'joint',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''


def test_diam_joint_fail():
    """Tests for diameter correction."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'repair', 'fail_diam.swc',
                             '-d', '2', '--diam', 'joint',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 1
    assert stdout == ''
    assert stderr == ''


def test_diam_sec():
    """Tests for diameter correction."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'repair', 'pass_simple_branch.swc',
                             '-d', '5', '--diam', 'sec',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''


def test_diam_order():
    """Tests for diameter correction."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'repair', 'pass_simple_branch.swc',
                             '-d', '5', '--diam', 'order',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''


def test_diam_order_pool():
    """Tests for diameter correction."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'repair', 'pass_simple_branch.swc',
                             '-d', '5', '--diam', 'order',
                             '--pool', 'pass_zjump.swc',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''


def test_diam_order_pool_err():
    """Tests for diameter correction."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'repair', 'pass_simple_branch.swc',
                             '-d', '5', '--diam', 'order',
                             '--pool', 'pass_soma.swc',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 1
    assert stdout == ''
    assert stderr == ''


def test_diam_breadth():
    """Tests for diameter correction."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'repair', 'pass_simple_branch.swc',
                             '-d', '5', '--diam', 'breadth',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''


def test_diam_breadth_pool():
    """Tests for diameter correction."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'repair', 'pass_simple_branch.swc',
                             '-d', '5', '--diam', 'breadth',
                             '--pool', 'pass_zjump.swc',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''


def test_diam_breadth_pool_err():
    """Tests for diameter correction."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'repair', 'pass_simple_branch.swc',
                             '-d', '5', '--diam', 'breadth',
                             '--pool', 'pass_soma.swc',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 1
    assert stdout == ''
    assert stderr == ''


def test_cut_repair():
    """Tests for repairing cut neurites."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'repair', 'pass_nmo_2_cut.swc',
                             '-c', '322', '341', '547', '1167',
                             '--seed', '1',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''


def test_cut_repair_alt():
    """Tests for repairing cut neurites."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'repair', 'pass_simple_branch.swc',
                             '-c', '11', '13', '--seed', '1',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''


def test_cut_repair_err():
    """Tests for repairing cut neurites."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'repair', 'pass_simple_branch.swc',
                             '-c', '7', '11', '13',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 3
    assert stdout == ''
    assert stderr == ''


def test_cut_repair_err_force():
    """Tests for repairing cut neurites."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'repair', 'pass_simple_branch.swc',
                             '-c', '7', '11', '13',
                             '--force-repair',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 3
    assert stdout == ''
    assert stderr == ''


def test_cut_repair_force():
    """Tests for repairing cut neurites."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'repair', 'pass_simple_branch_2.swc',
                             '-c', '11', '13', '14', '--seed', '1',
                             '--force-repair',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''


def test_cut_repair_pool():
    """Tests for repairing cut neurites."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'repair', 'pass_nmo_2_cut.swc',
                             '-c', '322', '341', '547', '1167',
                             '--pool', 'pass_nmo_1.swc',
                             '--seed', '1',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''


def test_cut_delete():
    """Tests for repairing cut neurites."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'repair', 'pass_simple_branch.swc',
                             '-l', '4', '8',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''


def test_cut_resample():
    """Tests for repairing cut neurites."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'repair', 'pass_simple_branch.swc',
                             '-r', '0.5',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''



"""Implementation of CLI view command."""

import matplotlib.pyplot as plt
import matplotlib as mpl

from cycler import cycler

from treem.morph import Morph
from treem.io import SWC

from treem.utils.plot import plot_neuron
from treem.utils.plot import plot_section
from treem.utils.plot import plot_tree
from treem.utils.plot import plot_points


_colors = ('coral', 'teal', 'darkgrey', 'royalblue', 'limegreen',
           'orchid', 'red', 'purple', 'orange', 'darkturquoise')
_NCOLORS = len(_colors)

mpl.rcParams['lines.linewidth'] = 1.0
mpl.rcParams['axes.prop_cycle'] = cycler(color=_colors)
mpl.rcParams.update({'font.size': 8})


def view(args):
    """Display neuron morphology structure."""
    # pylint: disable=too-many-locals
    # pylint: disable=too-many-branches
    # pylint: disable=too-many-statements
    # pylint: disable=expression-not-assigned
    fig = plt.figure(figsize=(8, 8))
    ax = fig.add_subplot(projection='3d')  # pylint: disable=invalid-name
    ax.xaxis.pane.set_edgecolor('w')
    ax.yaxis.pane.set_edgecolor('w')
    ax.zaxis.pane.set_edgecolor('w')
    ax.xaxis.pane.fill = False
    ax.yaxis.pane.fill = False
    ax.zaxis.pane.fill = False
    ax.set_proj_type('ortho')
    ax.set_xlabel('x')
    ax.set_ylabel('y')
    ax.set_zlabel('z')
    ax.grid(False)

    types = args.type if args.type else SWC.TYPES
    if args.title:
        fig.suptitle(args.title, fontsize=14)
    if args.no_axes:
        ax.set_axis_off()

    if args.mode == 'neurites':
        for file_name in reversed(args.file):
            morph = Morph(file_name)
            plot_neuron(ax, morph, types)
    elif args.mode == 'cells':
        for count, file_name in enumerate(reversed(args.file)):
            morph = Morph(file_name)
            colors = {k: f'C{count % _NCOLORS}' for k in types}
            plot_neuron(ax, morph, types, colors=colors)
    elif args.mode == 'shadow':
        for file_name in reversed(args.file[1:]):
            colors = {k: args.shadow_color for k in types}
            plot_neuron(ax, Morph(file_name), types, colors=colors,
                        linewidth=args.shadow_width)
        morph = Morph(args.file[0])
        plot_neuron(ax, morph, types)

    if args.branch:
        for group in args.branch:
            # pylint: disable=cell-var-from-loop
            # pylint: disable=undefined-loop-variable
            nodes = filter(lambda x: x.ident() in group, morph.root.walk())
            nodes = filter(lambda x: x.type() in types, nodes)
            for branch in nodes:
                plot_tree(ax, branch, morph.data, linewidth=1.5, color='C5')
                if args.show_id:
                    plot_points(ax, morph, group, types,
                                show_id=args.show_id)

    if args.sec:
        for group in args.sec:
            # pylint: disable=undefined-loop-variable
            nodes = filter(lambda x: x.ident() in group, morph.root.walk())
            nodes = filter(lambda x: x.type() in types, nodes)
            for sec in nodes:
                plot_section(ax, sec, morph.data, linewidth=1.5, color='C5')
                if args.show_id:
                    plot_points(ax, morph, group, types,
                                show_id=args.show_id)

    if args.mark:
        for group in args.mark:
            plot_points(ax, morph, group, types, show_id=args.show_id)

    if args.angle:
        ax.view_init(args.angle[0], args.angle[1])

    if args.proj:
        if args.proj.lower() == 'xy':
            ax.view_init(89.99, -90.01)
            ax.set_zlabel('')
            ax.set_zticks([])
        if args.proj.lower() == 'xz':
            ax.view_init(0.00, -90.01)
            ax.set_ylabel('')
            ax.set_yticks([])
        if args.proj.lower() == 'yz':
            ax.view_init(0.00, 0.01)
            ax.set_xlabel('')
            ax.set_xticks([])

    if args.dist:
        ax.dist = args.dist

    xmin = ax.xy_dataLim.xmin
    ymin = ax.xy_dataLim.ymin
    zmin = ax.zz_dataLim.xmin
    xmax = ax.xy_dataLim.xmax
    ymax = ax.xy_dataLim.ymax
    zmax = ax.zz_dataLim.xmax
    smax = max(max(ax.xy_dataLim.size), max(ax.zz_dataLim.size))
    ax.set_xlim((xmin+xmax-smax)/2, (xmin+xmax+smax)/2)
    ax.set_ylim((ymin+ymax-smax)/2, (ymin+ymax+smax)/2)
    ax.set_zlim((zmin+zmax-smax)/2, (zmin+zmax+smax)/2)

    if args.scale and args.scale > 0:
        ax.plot([xmax-args.scale, xmax], [ymin, ymin], [zmin, zmin],
                color='k', linewidth=3)
        ax.plot([xmax, xmax], [ymin, ymin+args.scale], [zmin, zmin],
                color='k', linewidth=3)
        ax.plot([xmax, xmax], [ymin, ymin], [zmin, zmin+args.scale],
                color='k', linewidth=3)

    plt.show() if not args.out else plt.savefig(args.out, dpi=100)



"""Testing CLI command measure."""

import subprocess
import os


def test_measure():
    """Tests for morphometric mesurements."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'measure', 'pass_simple_branch.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == """pass_simple_branch
dend area        24.0904
dend breadth           3
dend contrac           1
dend degree            2
dend diam          0.382
dend length      16.9706
dend nbranch           2
dend nstem             1
dend nterm             3
dend order             3
dend seclen      3.39411
dend volume      3.56734
dend xdim              8
dend ydim              5
dend zdim              0
soma area        12.5664
soma diam              2
soma volume      4.18879
soma xroot             0
soma yroot             0
soma zroot             0\n
"""
    assert stderr == ''


def test_out():
    """Tests for writing to file."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'measure', 'pass_simple_branch.swc',
                             '-o', '/tmp/test_treem.json'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''


def test_soma():
    """Tests for morphometric mesurements."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'measure', 'pass_soma.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout != ''
    assert stderr == ''



"""Testing CLI command modify."""

import subprocess
import os


def test_scale():
    """Tests for scaling of dimensions."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'modify', 'pass_simple_branch.swc',
                             '-s', '1', '1', '1', '-e', '2',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''


def test_scale_radius():
    """Tests for scaling of radii."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'modify', 'pass_simple_branch.swc',
                             '-r', '1', '-b', '1',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''


def test_unfold():
    """Tests for stretching and smoothing."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'modify', 'pass_zjump.swc',
                             '-t', '1', '-m', '1',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''


def test_jitter():
    """Tests for node jittering."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'modify', 'pass_simple_branch.swc',
                             '-i', '2', '4', '8', '-j', '0.3',
                             '--seed', '1',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''


def test_jitter_sec():
    """Tests for section jittering."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'modify', 'pass_simple_branch.swc',
                             '-i', '2', '4', '8', '-j', '0.3',
                             '--seed', '1', '--sec',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''


def test_twist():
    """Tests for branch twisting."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'modify', 'pass_simple_branch.swc',
                             '-i', '3', '9', '-w', '360',
                             '--seed', '1',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''


def test_swap():
    """Tests for branch swapping."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'modify', 'pass_simple_branch.swc',
                             '-i', '4', '8', '-a',
                             '--seed', '1',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''


def test_prune():
    """Tests for branch pruning."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'modify', 'pass_simple_branch.swc',
                             '-i', '4', '8', '-u',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    assert stdout == ''
    assert stderr == ''



"""Testing CLI command convert."""

import subprocess
import os


def test_unordered():
    """Tests for converting unordered SWC file."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'convert', 'fail_unordered.swc',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    _, stderr = proc.communicate()
    assert proc.returncode == 0
    # assert stdout == ''
    assert stderr == ''


def test_not_array():
    """Tests for converting incorrect file."""
    os.chdir(os.path.dirname(__file__) + '/data')
    proc = subprocess.Popen(['swc', 'convert', 'fail_not_array_1.swc',
                             '-o', '/tmp/test_treem.swc'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 1
    assert stdout == 'can not convert fail_not_array_1.swc.\n'
    assert stderr == ''



import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

dependencies = [
    "numpy",
    "matplotlib"
]

setuptools.setup(
    name="treem",
    version="1.0.0",
    author="Alexander Kozlov",
    author_email="akozlov@kth.se",
    description="Tree-like morphology data processing",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/a1eko/treem",
    packages=setuptools.find_packages(),
    install_requires=dependencies,
    entry_points={
        "console_scripts": [
            "swc = treem.cli:cli",
        ],
    },
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    keywords=[
        "neuron",
        "morphology",
        "reconstruction",
        "processing",
        "morphometry",
        "modification",
        "repair",
    ],
)



"""Testing module morph."""

import numpy as np

from treem import Node, Morph


def test_node_str():
    """Tests for Node string representation."""
    node = Node(value=1)
    assert str(node) == '1'


def test_node_point():
    """Tests for Node data point."""
    node = Node(value=np.array([1, 1, 0, 0, 0, 1, -1]))
    assert all(node.point() == np.array([0, 0, 0, 1]))


def test_morph_node():
    """Tests for Morph node lookup."""
    morph = Morph(data=np.array([[1, 1, 0, 0, 0, 1, -1]]))
    node = morph.node(1)
    assert node.ident() == 1


def test_morph_insert():
    """Tests for Morph node insertion."""
    morph = Morph(data=np.array([[1, 1, 0, 0, 0, 1, -1],
                                 [2, 1, 0, 0, 0, 1, 1]]))
    new_node = Node(value=np.array([0, 1, 0, 0, 0, 1, 0]))
    morph.insert(new_node, morph.node(2))
    assert morph.node(2).ident() == 2
    assert morph.node(3).ident() == 3
    assert morph.root.size() == 3



"""Import tests of the main module treem and it's submodules."""

# pylint: disable=unused-import
# pylint: disable=import-outside-toplevel


def test_import_treem():
    """Tests importing treem."""
    import treem


def test_import_node():
    """Tests importing Node from treem."""
    from treem import Node


def test_import_morph():
    """Tests importing Morph from treem."""
    from treem import Morph


def test_import_swc():
    """Tests importing SWC from treem."""
    from treem import SWC


def test_import_geom():
    """Tests importing geom utils from treem."""
    from treem.utils.geom import rotation_matrix
    from treem.utils.geom import norm
    from treem.utils.geom import angle_between
    from treem.utils.geom import rotation
    from treem.utils.geom import repair_branch
    from treem.utils.geom import sample


def test_import_plot():
    """Tests importing plot utils from treem."""
    from treem.utils.plot import plot_tree
    from treem.utils.plot import plot_section
    from treem.utils.plot import plot_neuron
    from treem.utils.plot import plot_points
